/* ipv4.cc
****************************************************************************************************************
****************************************************************************************************************

    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Askar Almatov

    This file is part of Libiclass. 
    
    Libiclass is free software: you can redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
    option) any later version.
    
    Libiclass is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.
    
    You should have received a copy of the GNU General Public License along with Libiclass. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <cstdlib>
#include <functional>
#include <limits>
#include <sstream>
#include <stdexcept>
#include <libiclass/ipv4>

using namespace iclass;
using namespace std;

/**************************************************************************************************************/
ipaddr4t::ipaddr4t( string Ipstr )
{
    uint32_t	       ipnum;
    string::size_type  tokenpos;
    string::size_type  tokenend;
    unsigned	       tokenvalue;

    if
    (
    	(tokenpos=Ipstr.find_first_not_of(" \t\n\v\f\r")) == string::npos ||
	(tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	(tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 255
    )
    {
    	throw invalid_argument( "ipaddr4t::ipaddr4t( std::string ) invalid w-octet" );
    }

    ipnum = (tokenvalue<<24);
    tokenpos = tokenend + 1;
    
    if
    (
    	tokenend == string::npos ||
    	Ipstr[tokenend] != '.' ||
	(tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	(tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 255
    )
    {
    	throw invalid_argument( "ipaddr4t::ipaddr4t( std::string ) invalid x-octet" );
    }

    ipnum += (tokenvalue<<16);
    tokenpos = tokenend + 1;

    if
    (
    	tokenend == string::npos ||
    	Ipstr[tokenend] != '.' ||
	//tokenpos >= Ipstr.size() ||
	(tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	(tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 255
    )
    {
    	throw invalid_argument( "ipaddr4t::ipaddr4t( std::string ) invalid y-octet" );
    }

    ipnum += (tokenvalue<<8);
    tokenpos = tokenend + 1;
    
    if
    (
    	tokenend == string::npos ||
    	Ipstr[tokenend] != '.' ||
	//tokenpos >= Ipstr.size() ||
	(tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	(tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 255 ||
	(tokenend!=string::npos && !isspace(Ipstr[tokenend]))
    )
    {
    	throw invalid_argument( "ipaddr4t::ipaddr4t( std::string ) invalid z-octet" );
    }

    ipnum += tokenvalue;
    _ipnum = ipnum;
}

/**************************************************************************************************************/
string
ipaddr4t::str() const
{
    ostringstream  oss;

    oss << woctet() << "." << xoctet() << "." << yoctet() << "." << zoctet();

    return oss.str();
}

/**************************************************************************************************************/
ipnet4t::ipnet4t() noexcept :
    // 0.0.0.0/0
    _frontip(),
    _backip( numeric_limits<uint32_t>::max() ),
    _nodebits( 32 )
{
}

/**************************************************************************************************************/
ipnet4t::ipnet4t( ipaddr4t Address ) noexcept :
    _frontip( Address ),
    _backip( Address ),
    _nodebits( 0 )
{
}

/**************************************************************************************************************/
ipnet4t::ipnet4t( ipaddr4t Address, ipaddr4t Mask ) :
    _frontip( Address )
{
    uint32_t  negmask( ~Mask );
    int       nodebits( 0 );

    while ( negmask & 1 )
    {
    	nodebits++;
	negmask >>= 1;
    }

    if ( negmask )
    {
	throw invalid_argument( "ipnet4t::ipnet4t( ipaddr4t, ipaddr4t ) invalid mask" ); 
    }
	
    if ( _frontip << (32-nodebits) )
    {
	throw invalid_argument( "ipnet4t::ipnet4t( ipaddr4t, ipaddr4t ) invalid address" ); 
    }
    
    _nodebits = nodebits;
    _backip = _frontip - 1 + ( (nodebits>=32)? 0 : (1<<nodebits) );
}

/**************************************************************************************************************/
ipnet4t::ipnet4t( string Ipstr )
{
    uint32_t	       ipnum;
    string::size_type  tokenpos;
    string::size_type  tokenend;
    unsigned	       tokenvalue;
    char	       endchar;

    if
    (
    	(tokenpos=Ipstr.find_first_not_of(" \t\n\v\f\r")) == string::npos ||
	(tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	(tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 255
    )
    {
    	throw invalid_argument( "ipnet4t::ipnet4t( std::string ) invalid w-octet" );
    }

    ipnum = (tokenvalue<<24);
    tokenpos = tokenend + 1;
    endchar = (tokenend==string::npos)? ' ' : Ipstr[tokenend];

    if ( !isspace(endchar) && endchar != '/' )
    {
    	if
    	(
    	    endchar != '.' ||
	    tokenpos >= Ipstr.size() ||
	    (tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	    (tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 255
    	)
    	{
    	    throw invalid_argument( "ipnet4t::ipnet4t( std::string ) invalid x-octet" );
    	}

    	ipnum += (tokenvalue<<16);
    	tokenpos = tokenend + 1;
    	endchar = (tokenend==string::npos)? ' ' : Ipstr[tokenend];

    	if ( !isspace(endchar) && endchar != '/' )
    	{
    	    if
    	    (
    	    	endchar != '.' ||
	    	tokenpos >= Ipstr.size() ||
	    	(tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	    	(tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 255
    	    )
    	    {
    	    	throw invalid_argument( "ipnet4t::ipnet4t( std::string ) invalid y-octet" );
    	    }

    	    ipnum += (tokenvalue<<8);
    	    tokenpos = tokenend + 1;
    	    endchar = (tokenend==string::npos)? ' ' : Ipstr[tokenend];

    	    if ( !isspace(endchar) && endchar != '/' )
    	    {
    	    	if
    	    	(
    	    	    endchar != '.' ||
	    	    tokenpos >= Ipstr.size() ||
	    	    (tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	    	    (tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 255
    	    	)
		{
    	   	    throw invalid_argument( "ipnet4t::ipnet4t( std::string ) invalid z-octet" );
    	   	}

    	   	ipnum += tokenvalue;
    	    	tokenpos = tokenend + 1;
    	    	endchar = (tokenend==string::npos)? ' ' : Ipstr[tokenend];
	    }
	}
    }
    
    if ( isspace(endchar) )
    {
    	_frontip = _backip = ipnum;
    	_nodebits = 0;

	return;
    }

    if
    (
    	endchar != '/' ||
	tokenpos >= Ipstr.size() ||
	(tokenend=Ipstr.find_first_not_of("0123456789",tokenpos)) == tokenpos ||
	(tokenvalue=strtol(Ipstr.substr(tokenpos,tokenend-tokenpos).c_str(),0,10)) > 32 ||
	(tokenend!=string::npos && !isspace(Ipstr[tokenend]))
    )
    {
    	throw invalid_argument( "ipnet4t::ipnet4t( std::string ) invalid network bits");
    }

    if ( tokenvalue < 32 && (ipnum<<tokenvalue) )
    {
	throw invalid_argument( "ipnet4t::ipnet4t( std::string ) invalid address" );
    }
    
    _nodebits = 32 - tokenvalue;
    _frontip = ipnum;
    _backip = ipnum - 1 + ( (!tokenvalue)? 0 : (1<<_nodebits) );
}

/**************************************************************************************************************/
ipnet4t&
ipnet4t::merge()
{
    if ( _nodebits >= 32 )
    {
	throw overflow_error( "ipnet4t::merge() node bits overflow" );
    }

    _frontip = _frontip & ~(1<<_nodebits++);
    _backip = _frontip - 1 + ( (_nodebits>=32)? 0 : (1<<_nodebits) );

    return *this;
}

/**************************************************************************************************************/
string
ipnet4t::str() const
{
    ostringstream  oss;

    oss << _frontip.woctet();

    if ( _nodebits < 24 )
    {
	oss << "." << _frontip.xoctet();
	
    	if ( _nodebits < 16 )
    	{
	    oss << "." << _frontip.yoctet();
	    
	    if ( _nodebits < 8 )
	    {
		oss << "." << _frontip.zoctet();
	    }
    	}
    }
    
    if ( _nodebits > 0 )
    {
    	oss << "/" << (32 - _nodebits);
    }

    return oss.str();
}

/**************************************************************************************************************/
void
ipset4t::insert( ipnet4t Net )
{
    set<ipnet4t>::iterator lb( lower_bound(_networks.begin(), _networks.end(), Net) );

    while ( lb != _networks.end() )
    {
	if ( Net.contains(*lb) )
	{
	    _networks.erase( lb++ );
	}
	else if ( Net.ismergeable(*lb) )
	{
	    Net.merge();
	    _networks.erase( lb++ );
	}
	else
	{
	    break;
	}
    }

    while ( lb != _networks.begin() )
    {
	if ( (--lb)->contains(Net) )
	{
	    return;
	}
	else if ( lb->ismergeable(Net) )
	{
	    Net.merge();
	    _networks.erase( lb++ );

	    while ( lb != _networks.end() && Net.ismergeable(*lb) )
	    {
		Net.merge();
		_networks.erase( lb++ );
	    }
	}
	else
	{
	    lb++;

	    break;
	}
    }

    _networks.insert( lb, Net );
}

/**************************************************************************************************************/
iplookup4t::iplookup4t( const ipset4t& Set, int Pfxbits ) :
    _pfxbits( Pfxbits ),
    _table( NULL )
{
    if ( _pfxbits <= 0  ||  _pfxbits > 20 )
    {
    	throw range_error( "iplookup4t::iplookup4t( ... ) prefix bits is out of range");
    }

    _table = new std::list<ipnet4t>[ 1 << _pfxbits ];

    for ( const auto& net : Set )
    {
    	list<ipnet4t>*  plist( _table + (net.address() >> (32-_pfxbits)) );

	if ( net.networkbits() >= _pfxbits )
	{
	    plist->push_back( net );
	}
	else
	{
	    for ( int repeat = (1 << (_pfxbits - net.networkbits())); repeat > 0; --repeat )
	    {
		(plist++)->push_back( net );
	    }
	}
    }
}

/**************************************************************************************************************/
iplookup4t::~iplookup4t()
{
    if ( _table )
    {
    	delete[] _table;
    }
}

/**************************************************************************************************************/
bool
iplookup4t::contains( ipaddr4t Ip ) const
{
    list<ipnet4t>*  plist( _table + (Ip >> (32-_pfxbits)) );

    return
    (
	!plist->empty() &&

        find_if
        (
            plist->begin(),
            plist->end(),
            bind2nd( mem_fun_ref((bool (ipnet4t::*)(ipaddr4t) const)&ipnet4t::contains), Ip )
        )
        != plist->end()
    );
}

/**************************************************************************************************************/
bool
iplookup4t::contains( const ipnet4t& Net ) const
{
    list<ipnet4t>*  plist( _table + (Net.address() >> (32-_pfxbits)) );

    return
    (
	!plist->empty() &&

        find_if
        (
            plist->begin(),
            plist->end(),
            bind2nd( mem_fun_ref((bool (ipnet4t::*)(ipnet4t) const)&ipnet4t::contains), Net )
        )
        != plist->end()
    );
}
