/* task.cc
****************************************************************************************************************
****************************************************************************************************************

    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Askar Almatov

    This file is part of Libiclass. 
    
    Libiclass is free software: you can redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
    option) any later version.
    
    Libiclass is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.
    
    You should have received a copy of the GNU General Public License along with Libiclass. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include <cassert>
#include <libiclass/task>

#if defined( __FreeBSD__ )
#include <pthread_np.h>
#else
#include <unistd.h>
#include <sys/syscall.h>
#endif

using namespace iclass;
using namespace std;

/**************************************************************************************************************/
static int
_ThreadId()
{

#if defined( __FreeBSD__ )

    return pthread_getthreadid_np();

#elif defined( __MACH__ )

    return syscall( SYS_thread_selfid );

#else

    return syscall( SYS_gettid );

#endif

}

/**************************************************************************************************************/
priorityt::priorityt()
{
    struct sched_param	scparam;
    int			scpolicy;

    assert( pthread_getschedparam(pthread_self(), &scpolicy, &scparam) == 0 );
    _policy = scpolicy;
    _value = scparam.sched_priority;
}

/**************************************************************************************************************/
priorityt::priorityt( int Policy, int Value )
{
    int  min( sched_get_priority_min(Policy) );
    int  max( sched_get_priority_max(Policy) );

    assert( min != -1 );			// is policy supported
    _policy = Policy;
    assert( Value >= min && Value <= max );
    _value = Value;
}

/**************************************************************************************************************/
void
priorityt::policy( int Policy )
{
    int  min( sched_get_priority_min(Policy) );
    int  max( sched_get_priority_max(Policy) );

    assert( min != -1 );			// is policy supported
    _policy = Policy;
    assert( _value >= min && _value <= max );
}

/**************************************************************************************************************/
void
priorityt::value( int Value )
{
    assert( Value >= minvalue() && Value <= maxvalue() );
    _value = Value;
}

/**************************************************************************************************************/
int
priorityt::operator -( const priorityt& Rhs ) const
{
    assert( _policy == Rhs.policy() );

    return _value - Rhs.value();
}

/**************************************************************************************************************/
string
priorityt::policyname() const
{
    switch ( _policy )
    {
	case SCHED_OTHER:

	    return "SCHED_OTHER";

	case SCHED_RR:

	    return "SCHED_RR";

	case SCHED_FIFO:

	    return "SCHED_FIFO";
    };

    return to_string( _policy );
}

/**************************************************************************************************************/
int
priorityt::minvalue() const
{
    return sched_get_priority_min( _policy );
}

/**************************************************************************************************************/
int
priorityt::maxvalue() const
{
    return sched_get_priority_max( _policy );
}

/**************************************************************************************************************/
priorityt
priorityt::min( int Policy )
{
    int  min( sched_get_priority_min(Policy) );

    assert( min != -1 );			// is policy supported?

    return priorityt( Policy, min );
}

/**************************************************************************************************************/
priorityt
priorityt::max( int Policy )
{
    int  max( sched_get_priority_max(Policy) );

    assert( max != -1 );			// is policy supported?

    return priorityt( Policy, max );
}

/**************************************************************************************************************/
threadt::threadt()
{
    __td = pthread_self();
}

/**************************************************************************************************************/
threadt::threadt( taskt* Ptask, priorityt Prio )
{
    pthread_attr_t	attr;
    struct sched_param	scparam;

    scparam.sched_priority = Prio.value();
    assert( pthread_attr_init(&attr) == 0 );
    assert( pthread_attr_setschedpolicy(&attr, Prio.policy()) == 0 );
    assert( pthread_attr_setschedparam(&attr, &scparam) == 0 );
    assert( pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED) == 0 );
    assert( pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM) == 0 );
    assert( pthread_create(&__td, &attr, taskt::__routine, Ptask) == 0 );
    assert( pthread_attr_destroy(&attr) == 0 );
}

/**************************************************************************************************************/
threadt::~threadt()
{
}

/**************************************************************************************************************/
void
threadt::waitend()
/*
    wait until thread will be executed
*/
{
    assert( pthread_join(__td, NULL) == 0 );
}

/**************************************************************************************************************/
void
threadt::prio( priorityt Prio )
{
    struct sched_param	scparam;
    int			scpolicy;

    assert( pthread_getschedparam(__td, &scpolicy, &scparam) == 0 );
    scparam.sched_priority = Prio.value();
    assert( pthread_setschedparam(__td, Prio.policy(), &scparam) == 0 );
}

/**************************************************************************************************************/
priorityt
threadt::prio() const
{
    struct sched_param	scparam;
    int			scpolicy;

    assert( pthread_getschedparam(__td, &scpolicy, &scparam) == 0 );

    return priorityt( scpolicy, scparam.sched_priority );
}

/**************************************************************************************************************/
taskt::taskt() :
    __pthread( nullptr ),
    __a_tid( -1 ),
    __a_started( false ),
    __a_stopped( false ),
    __a_finished( false )
{
}

/**************************************************************************************************************/
taskt::~taskt()
{
    if ( __pthread )
    {
	delete __pthread;
    }
}

/**************************************************************************************************************/
void
taskt::start( priorityt Prio )
{
    if ( !__a_started )
    {
	__a_started = true;
        __pthread = new threadt( this, Prio );
    }
}

/**************************************************************************************************************/
void
taskt::stop()
{
    if ( !__a_stopped && __a_started )
    {
	__a_stopped = true;

	if ( !__a_finished )
	{
	    __pthread->waitend();
	}
    }
}

/**************************************************************************************************************/
void*            
taskt::__routine( void* Ptask )
{
    taskt*  ptask( reinterpret_cast<taskt*>(Ptask) );

    ptask->__a_tid = _ThreadId();
    ptask->routine();
    ptask->__a_finished = true;
    pthread_exit( NULL );

    return 0;
}

