/* time.cc
****************************************************************************************************************
****************************************************************************************************************

    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Askar Almatov

    This file is part of Libiclass. 
    
    Libiclass is free software: you can redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
    option) any later version.
    
    Libiclass is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.
    
    You should have received a copy of the GNU General Public License along with Libiclass. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include <cassert>
#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <libiclass/time>

using namespace iclass;
using namespace std;

/*
    Компоненты в порядке увеличения старшинства:
        секунды,
        минуты,
        часы,
        день месяца, день, день года, день недели
        месяц,
        год.
    Базовые компоненты: секунды, минуты, часы, день месяца, месяц, год.
    Установка базовых компонент не должна изменять младшие базовые компоненты,
    за исключением случаев когда результат получается не корректным (distortion), а именно
    при некорректном дне месяца после установки года или месяца.
    Пусть время 31-янв, установим месяц февраль. Как интерпретировать 31-фев?
    Наверное надо преобразовать в 28-фев или 29-фев для високосного года.
*/
/* 
    TODO
    
    1. Должна иметься стандартная таблица високосных секунд - первоисточник. За образец может быть взята таблица
	    /usr/src/glibc-x.x.x/timezone/leapseconds

    2. Должен иметься компилятор
	    leapseconds ==> LSTBL

    3. Должна иметься тестовая функция для проверки работы класса с високосными секундами. Проверка
       корректности должна производиться по первоисточнику, т.е. таблице leapseconds.

    4. Должен иметься глобальный метод подгрузки новых значений таблицы LSTBL из файла. Это необходимо
       чтобы при появлении новых високосных секунд не приходилось выпускать новый релиз библиотеки классов,
       а достаточно было бы разослать новую таблицу високосных секунд, а еще лучше, если бы она подгружалась
       из Интернета. Надо определиться с форматом этого файла:
    	    - .h файл
	    - .cpp файл
	    - leapseconds файл (недостаток: возможны ошибки при компиляции)
	    - просто таблица
*/

extern "C" time_t	time2posix( time_t );

const gtimet::wdayt     BASEWEEKDAY         	     	= gtimet::SAT; 		// 2000-01-01 00:00:00 UTC
const long double	BASEJULIANDATE			= 2451544.5L;		// 2000-01-01 00:00:00 TT
const long double	BASEMODIFIEDJULIANDATE		= 51544.0L;		// 2000-01-01 00:00:00 TT

const long double	HYEPSILON			= 0.01095163L;		// half year suburb radius
const long double	TMSLTTDELTA			= 64.184L;		// TMSL = TT - TMSLTTDELTA
const long double	TMSLTAIDELTA			= 32.0L;		// TMSL = TAI - TMSLTAIDELTA
const long double	TMSLGPSDELTA			= 13.0L;		// TMSL = GPS - TMSLTAIDELTA
const long double       TMS19700101 	 	    	= -946684800.0L;	// 1970-01-01 00:00:00 UTC _tms value
const long double       TMSL19700101 	     		= -946684822.0L;	// 1970-01-01 00:00:00 UTC _tmsl value

const long double	gtimet::ATOMICYEAR           	= 31556925.9747L; 	// duration of year 1900
const long double	gtimet::GRIGORIANYEAR          	= 31556952.0L;
const long double	gtimet::JULIANYEAR           	= 31557600.0L;

const long double       gtimet::SECOND               	= 1.0L;
const long double       gtimet::MEANMINUTE     	     	= 60.0L;
const long double       gtimet::MEANHOUR      	     	= 3600.0L;
const long double       gtimet::MEANDAY      	     	= 86400.0L;
const long double       gtimet::MEANMONTH 	     	= gtimet::GRIGORIANYEAR/12;
const long double       gtimet::MEANYEAR 	     	= gtimet::GRIGORIANYEAR;

struct lsct						// leap second candidate type
{
    int			lscn;				// leap second candidate number
    int			lsn;				// leap second number
    long double		lstmsl;				// begin of last june|december second linear timestamp
    long double		lstms;				// begin of last june|december second timestamp (not linear)
    long double		lsyear;				// begin of last june|december second UTC year field value
    int			lsmonth;			// begin of last june|december second UTC month field value
    int			lsmday;				// begin of last june|december second UTC month day field value
    int			lshour;				// begin of last june|december second UTC hour field value
    int			lsminute;			// begin of last june|december second UTC minute field value
    int			lssecond;			// begin of last june|december second UTC second field value
};

const int		LSTBLNUM = 82;
const lsct		LSTBL[] = 
{
    { 0,   0, -867888022.0L, -867888000.0L, 1972.0L,  6-1, 30-1, 23, 59, 60 },
    { 1,   1, -851990421.0L, -851990400.0L, 1972.0L, 12-1, 31-1, 23, 59, 60 },
    { 2,  -1, -836352021.0L, -836352001.0L, 1973.0L,  6-1, 30-1, 23, 59, 59 },
    { 3,   3, -820454420.0L, -820454400.0L, 1973.0L, 12-1, 31-1, 23, 59, 60 },
    { 4,  -1, -804816020.0L, -804816001.0L, 1974.0L,  6-1, 30-1, 23, 59, 59 },
    { 5,   5, -788918419.0L, -788918400.0L, 1974.0L, 12-1, 31-1, 23, 59, 60 },
    { 6,  -1, -773280019.0L, -773280001.0L, 1975.0L,  6-1, 30-1, 23, 59, 59 },
    { 7,   7, -757382418.0L, -757382400.0L, 1975.0L, 12-1, 31-1, 23, 59, 60 },
    { 8,  -1, -741657618.0L, -741657601.0L, 1976.0L,  6-1, 30-1, 23, 59, 59 },
    { 9,   9, -725760017.0L, -725760000.0L, 1976.0L, 12-1, 31-1, 23, 59, 60 },
    { 10, -1, -710121617.0L, -710121601.0L, 1977.0L,  6-1, 30-1, 23, 59, 59 },
    { 11, 11, -694224016.0L, -694224000.0L, 1977.0L, 12-1, 31-1, 23, 59, 60 },
    { 12, -1, -678585616.0L, -678585601.0L, 1978.0L,  6-1, 30-1, 23, 59, 59 },
    { 13, 13, -662688015.0L, -662688000.0L, 1978.0L, 12-1, 31-1, 23, 59, 60 },
    { 14, -1, -647049615.0L, -647049601.0L, 1979.0L,  6-1, 30-1, 23, 59, 59 },
    { 15, 15, -631152014.0L, -631152000.0L, 1979.0L, 12-1, 31-1, 23, 59, 60 },
    { 16, -1, -615427214.0L, -615427201.0L, 1980.0L,  6-1, 30-1, 23, 59, 59 },
    { 17, -1, -599529614.0L, -599529601.0L, 1980.0L, 12-1, 31-1, 23, 59, 59 },
    { 18, 18, -583891213.0L, -583891200.0L, 1981.0L,  6-1, 30-1, 23, 59, 60 },
    { 19, -1, -567993613.0L, -567993601.0L, 1981.0L, 12-1, 31-1, 23, 59, 59 },
    { 20, 20, -552355212.0L, -552355200.0L, 1982.0L,  6-1, 30-1, 23, 59, 60 },
    { 21, -1, -536457612.0L, -536457601.0L, 1982.0L, 12-1, 31-1, 23, 59, 59 },
    { 22, 22, -520819211.0L, -520819200.0L, 1983.0L,  6-1, 30-1, 23, 59, 60 },
    { 23, -1, -504921611.0L, -504921601.0L, 1983.0L, 12-1, 31-1, 23, 59, 59 },
    { 24, -1, -489196811.0L, -489196801.0L, 1984.0L,  6-1, 30-1, 23, 59, 59 },
    { 25, -1, -473299211.0L, -473299201.0L, 1984.0L, 12-1, 31-1, 23, 59, 59 },
    { 26, 26, -457660810.0L, -457660800.0L, 1985.0L,  6-1, 30-1, 23, 59, 60 },
    { 27, -1, -441763210.0L, -441763201.0L, 1985.0L, 12-1, 31-1, 23, 59, 59 },
    { 28, -1, -426124810.0L, -426124801.0L, 1986.0L,  6-1, 30-1, 23, 59, 59 },
    { 29, -1, -410227210.0L, -410227201.0L, 1986.0L, 12-1, 31-1, 23, 59, 59 },
    { 30, -1, -394588810.0L, -394588801.0L, 1987.0L,  6-1, 30-1, 23, 59, 59 },
    { 31, 31, -378691209.0L, -378691200.0L, 1987.0L, 12-1, 31-1, 23, 59, 60 },
    { 32, -1, -362966409.0L, -362966401.0L, 1988.0L,  6-1, 30-1, 23, 59, 59 },
    { 33, -1, -347068809.0L, -347068801.0L, 1988.0L, 12-1, 31-1, 23, 59, 59 },
    { 34, -1, -331430409.0L, -331430401.0L, 1989.0L,  6-1, 30-1, 23, 59, 59 },
    { 35, 35, -315532808.0L, -315532800.0L, 1989.0L, 12-1, 31-1, 23, 59, 60 },
    { 36, -1, -299894408.0L, -299894401.0L, 1990.0L,  6-1, 30-1, 23, 59, 59 },
    { 37, 37, -283996807.0L, -283996800.0L, 1990.0L, 12-1, 31-1, 23, 59, 60 },
    { 38, -1, -268358407.0L, -268358401.0L, 1991.0L,  6-1, 30-1, 23, 59, 59 },
    { 39, -1, -252460807.0L, -252460801.0L, 1991.0L, 12-1, 31-1, 23, 59, 59 },
    { 40, 40, -236736006.0L, -236736000.0L, 1992.0L,  6-1, 30-1, 23, 59, 60 },
    { 41, -1, -220838406.0L, -220838401.0L, 1992.0L, 12-1, 31-1, 23, 59, 59 },
    { 42, 42, -205200005.0L, -205200000.0L, 1993.0L,  6-1, 30-1, 23, 59, 60 },
    { 43, -1, -189302405.0L, -189302401.0L, 1993.0L, 12-1, 31-1, 23, 59, 59 },
    { 44, 44, -173664004.0L, -173664000.0L, 1994.0L,  6-1, 30-1, 23, 59, 60 },
    { 45, -1, -157766404.0L, -157766401.0L, 1994.0L, 12-1, 31-1, 23, 59, 59 },
    { 46, -1, -142128004.0L, -142128001.0L, 1995.0L,  6-1, 30-1, 23, 59, 59 },
    { 47, 47, -126230403.0L, -126230400.0L, 1995.0L, 12-1, 31-1, 23, 59, 60 },
    { 48, -1, -110505603.0L, -110505601.0L, 1996.0L,  6-1, 30-1, 23, 59, 59 },
    { 49, -1,  -94608003.0L,  -94608001.0L, 1996.0L, 12-1, 31-1, 23, 59, 59 },
    { 50, 50,  -78969602.0L,  -78969600.0L, 1997.0L,  6-1, 30-1, 23, 59, 60 },
    { 51, -1,  -63072002.0L,  -63072001.0L, 1997.0L, 12-1, 31-1, 23, 59, 59 },
    { 52, -1,  -47433602.0L,  -47433601.0L, 1998.0L,  6-1, 30-1, 23, 59, 59 },
    { 53, 53,  -31536001.0L,  -31536000.0L, 1998.0L, 12-1, 31-1, 23, 59, 60 },
    { 54, -1,  -15897601.0L,  -15897601.0L, 1999.0L,  6-1, 30-1, 23, 59, 59 },
    { 55, -1,         -1.0L,         -1.0L, 1999.0L, 12-1, 31-1, 23, 59, 59 },
    { 56, -1,   15724799.0L,   15724799.0L, 2000.0L,  6-1, 30-1, 23, 59, 59 },
    { 57, -1,   31622399.0L,   31622399.0L, 2000.0L, 12-1, 31-1, 23, 59, 59 },
    { 58, -1,   47260799.0L,   47260799.0L, 2001.0L,  6-1, 30-1, 23, 59, 59 },
    { 59, -1,   63158399.0L,   63158399.0L, 2001.0L, 12-1, 31-1, 23, 59, 59 },
    { 60, -1,   78796799.0L,   78796799.0L, 2002.0L,  6-1, 30-1, 23, 59, 59 },
    { 61, -1,   94694399.0L,   94694399.0L, 2002.0L, 12-1, 31-1, 23, 59, 59 },
    { 62, -1,  110332799.0L,  110332799.0L, 2003.0L,  6-1, 30-1, 23, 59, 59 },
    { 63, -1,  126230399.0L,  126230399.0L, 2003.0L, 12-1, 31-1, 23, 59, 59 },
    { 64, -1,  141955199.0L,  141955199.0L, 2004.0L,  6-1, 30-1, 23, 59, 59 },
    { 65, -1,  157852799.0L,  157852799.0L, 2004.0L, 12-1, 31-1, 23, 59, 59 },
    { 66, -1,  173491199.0L,  173491199.0L, 2005.0L,  6-1, 30-1, 23, 59, 59 },
    { 67, 67,  189388800.0L,  189388800.0L, 2005.0L, 12-1, 31-1, 23, 59, 60 },
    { 68, -1,  205027200.0L,  205027199.0L, 2006.0L,  6-1, 30-1, 23, 59, 59 },
    { 69, -1,  220924800.0L,  220924799.0L, 2006.0L, 12-1, 31-1, 23, 59, 59 },
    { 70, -1,  236563200.0L,  236563199.0L, 2007.0L,  6-1, 30-1, 23, 59, 59 },
    { 71, -1,  252460800.0L,  252460799.0L, 2007.0L, 12-1, 31-1, 23, 59, 59 },
    { 72, -1,  268185600.0L,  268185599.0L, 2008.0L,  6-1, 30-1, 23, 59, 59 },
    { 73, 73,  284083201.0L,  284083200.0L, 2008.0L, 12-1, 31-1, 23, 59, 60 },
    { 74, -1,  299721601.0L,  299721599.0L, 2009.0L,  6-1, 30-1, 23, 59, 59 },
    { 75, -1,  315619201.0L,  315619199.0L, 2009.0L, 12-1, 31-1, 23, 59, 59 },
    { 76, -1,  331257601.0L,  331257599.0L, 2010.0L,  6-1, 30-1, 23, 59, 59 },
    { 77, -1,  347155201.0L,  347155199.0L, 2010.0L, 12-1, 31-1, 23, 59, 59 },
    { 78, -1,  362793601.0L,  362793599.0L, 2011.0L,  6-1, 30-1, 23, 59, 59 },
    { 79, -1,  378691201.0L,  378691199.0L, 2011.0L, 12-1, 31-1, 23, 59, 59 },
    { 80, 80,  394416002.0L,  394416000.0L, 2012.0L,  6-1, 30-1, 23, 59, 60 },
    { 81, -1,  410313602.0L,  410313599.0L, 2012.0L, 12-1, 31-1, 23, 59, 59 }
};

const long double	DAYSMAGICTBL[] =
{
     0.1425L,  0.1850L,  0.0275L,  0.0700L,  0.1125L,  0.1550L, -0.0025L,  0.0400L,  0.0825L,  0.1250L, 
    -0.0325L,  0.0100L,  0.0525L,  0.0950L, -0.0625L, -0.0200L,  0.0225L,  0.0650L, -0.0925L, -0.0500L, 
    -0.0075L,  0.0350L, -0.1225L, -0.0800L, -0.0375L,  0.0050L, -0.1525L, -0.1100L, -0.0675L, -0.0250L, 
    -0.1825L, -0.1400L, -0.0975L, -0.0550L, -0.2125L, -0.1700L, -0.1275L, -0.0850L, -0.2425L, -0.2000L, 
    -0.1575L, -0.1150L, -0.2725L, -0.2300L, -0.1875L, -0.1450L, -0.3025L, -0.2600L, -0.2175L, -0.1750L, 
    -0.3325L, -0.2900L, -0.2475L, -0.2050L, -0.3625L, -0.3200L, -0.2775L, -0.2350L, -0.3925L, -0.3500L, 
    -0.3075L, -0.2650L, -0.4225L, -0.3800L, -0.3375L, -0.2950L, -0.4525L, -0.4100L, -0.3675L, -0.3250L, 
    -0.4825L, -0.4400L, -0.3975L, -0.3550L, -0.5125L, -0.4700L, -0.4275L, -0.3850L, -0.5425L, -0.5000L, 
    -0.4575L, -0.4150L, -0.5725L, -0.5300L, -0.4875L, -0.4450L, -0.6025L, -0.5600L, -0.5175L, -0.4750L, 
    -0.6325L, -0.5900L, -0.5475L, -0.5050L, -0.6625L, -0.6200L, -0.5775L, -0.5350L, -0.4925L, -0.2500L, 
    -0.0075L,  0.2350L,  0.2775L,  0.3200L,  0.3625L,  0.4050L,  0.2475L,  0.2900L,  0.3325L,  0.3750L, 
     0.2175L,  0.2600L,  0.3025L,  0.3450L,  0.1875L,  0.2300L,  0.2725L,  0.3150L,  0.1575L,  0.2000L, 
     0.2425L,  0.2850L,  0.1275L,  0.1700L,  0.2125L,  0.2550L,  0.0975L,  0.1400L,  0.1825L,  0.2250L, 
     0.0675L,  0.1100L,  0.1525L,  0.1950L,  0.0375L,  0.0800L,  0.1225L,  0.1650L,  0.0075L,  0.0500L, 
     0.0925L,  0.1350L, -0.0225L,  0.0200L,  0.0625L,  0.1050L, -0.0525L, -0.0100L,  0.0325L,  0.0750L, 
    -0.0825L, -0.0400L,  0.0025L,  0.0450L, -0.1125L, -0.0700L, -0.0275L,  0.0150L, -0.1425L, -0.1000L, 
    -0.0575L, -0.0150L, -0.1725L, -0.1300L, -0.0875L, -0.0450L, -0.2025L, -0.1600L, -0.1175L, -0.0750L, 
    -0.2325L, -0.1900L, -0.1475L, -0.1050L, -0.2625L, -0.2200L, -0.1775L, -0.1350L, -0.2925L, -0.2500L, 
    -0.2075L, -0.1650L, -0.3225L, -0.2800L, -0.2375L, -0.1950L, -0.3525L, -0.3100L, -0.2675L, -0.2250L, 
    -0.3825L, -0.3400L, -0.2975L, -0.2550L, -0.4125L, -0.3700L, -0.3275L, -0.2850L, -0.2425L,  0.0000L, 
     0.2425L,  0.4850L,  0.5275L,  0.5700L,  0.6125L,  0.6550L,  0.4975L,  0.5400L,  0.5825L,  0.6250L, 
     0.4675L,  0.5100L,  0.5525L,  0.5950L,  0.4375L,  0.4800L,  0.5225L,  0.5650L,  0.4075L,  0.4500L, 
     0.4925L,  0.5350L,  0.3775L,  0.4200L,  0.4625L,  0.5050L,  0.3475L,  0.3900L,  0.4325L,  0.4750L, 
     0.3175L,  0.3600L,  0.4025L,  0.4450L,  0.2875L,  0.3300L,  0.3725L,  0.4150L,  0.2575L,  0.3000L, 
     0.3425L,  0.3850L,  0.2275L,  0.2700L,  0.3125L,  0.3550L,  0.1975L,  0.2400L,  0.2825L,  0.3250L, 
     0.1675L,  0.2100L,  0.2525L,  0.2950L,  0.1375L,  0.1800L,  0.2225L,  0.2650L,  0.1075L,  0.1500L, 
     0.1925L,  0.2350L,  0.0775L,  0.1200L,  0.1625L,  0.2050L,  0.0475L,  0.0900L,  0.1325L,  0.1750L, 
     0.0175L,  0.0600L,  0.1025L,  0.1450L, -0.0125L,  0.0300L,  0.0725L,  0.1150L, -0.0425L, -0.0000L, 
     0.0425L,  0.0850L, -0.0725L, -0.0300L,  0.0125L,  0.0550L, -0.1025L, -0.0600L, -0.0175L,  0.0250L, 
    -0.1325L, -0.0900L, -0.0475L, -0.0050L, -0.1625L, -0.1200L, -0.0775L, -0.0350L,  0.0075L,  0.2500L, 
     0.4925L,  0.7350L,  0.7775L,  0.8200L,  0.8625L,  0.9050L,  0.7475L,  0.7900L,  0.8325L,  0.8750L, 
     0.7175L,  0.7600L,  0.8025L,  0.8450L,  0.6875L,  0.7300L,  0.7725L,  0.8150L,  0.6575L,  0.7000L, 
     0.7425L,  0.7850L,  0.6275L,  0.6700L,  0.7125L,  0.7550L,  0.5975L,  0.6400L,  0.6825L,  0.7250L, 
     0.5675L,  0.6100L,  0.6525L,  0.6950L,  0.5375L,  0.5800L,  0.6225L,  0.6650L,  0.5075L,  0.5500L, 
     0.5925L,  0.6350L,  0.4775L,  0.5200L,  0.5625L,  0.6050L,  0.4475L,  0.4900L,  0.5325L,  0.5750L, 
     0.4175L,  0.4600L,  0.5025L,  0.5450L,  0.3875L,  0.4300L,  0.4725L,  0.5150L,  0.3575L,  0.4000L, 
     0.4425L,  0.4850L,  0.3275L,  0.3700L,  0.4125L,  0.4550L,  0.2975L,  0.3400L,  0.3825L,  0.4250L, 
     0.2675L,  0.3100L,  0.3525L,  0.3950L,  0.2375L,  0.2800L,  0.3225L,  0.3650L,  0.2075L,  0.2500L, 
     0.2925L,  0.3350L,  0.1775L,  0.2200L,  0.2625L,  0.3050L,  0.1475L,  0.1900L,  0.2325L,  0.2750L, 
     0.1175L,  0.1600L,  0.2025L,  0.2450L,  0.0875L,  0.1300L,  0.1725L,  0.2150L,  0.0575L,  0.1000L
};

/**************************************************************************************************************/
static int
_Irand()
/*
    Uses a multiplicative congruential random number generator with period 2^32 to return successive pseudo-
    random numbers in the range from 0 to 2^15 - 1.
*/
{
    static long  seed1 = 1;
    static long  seed2 = 1;

    seed1 = 22695477L * seed1 + 1;
    seed2 = 31415926L * seed2 + 7;

    return ( int(seed1>>16) ^ int(seed2>>16) ) & 0x7fff;
}

/**************************************************************************************************************/
static double
_Drand0()
/*
    [0;1[ равномерно распределенная величина
*/
{
    return double( rand() ) / ( double(RAND_MAX) + 1 );
}

/**************************************************************************************************************/
static double
_Drand1()
/*
    ]0;1] равномерно распределенная величина
*/
{
    int  randint;
    
    for ( ; !(randint = rand()); ) ;
    
    return double( randint ) / RAND_MAX;
}

/**************************************************************************************************************/
static double
_NormRand()
/*
    нормально распределенная величина M == 0, D == 1
*/
{
    double  v1, v2, s;

    for ( ;; )
    {
        v1 = _Drand0() * 2 - 1;
        v2 = _Drand0() * 2 - 1;
        s  = v1 * v1 + v2 * v2;

        if ( s < 1 )
	{
	    break;
	}
    }
    
    return v2 * sqrt( -2 * log(s) / s );
}

/**************************************************************************************************************/
gtimet::gtimet( time_t Unixtime, lsmodet Lsmode ) :
    _tmsl( 0.0L ),
    _ptzi( 0 ),
    _readyflags( 0 ),
    _distortionflags( 0 )
{
    setunixtime( Unixtime, Lsmode );
}

/**************************************************************************************************************/
gtimet::gtimet( const struct timeval& Tval, lsmodet Lsmode ) :
    _tmsl( 0.0L ),
    _ptzi( 0 ),
    _readyflags( 0 ),
    _distortionflags( 0 )
{
    setunixtime( Tval.tv_sec, Lsmode );
    _tmsl += static_cast<long double>( Tval.tv_usec ) / 1000000;
}

/**************************************************************************************************************/
gtimet::gtimet( const string& TimeString, string* Perrdesc ) :
    _tmsl( 0.0L ),
    _ptzi( 0 ),
    _readyflags( 0 ),
    _distortionflags( 0 )
{
    setall( TimeString, Perrdesc );
}

/**************************************************************************************************************/
gtimet::gtimet( long double Y, int Mo, int D, int H, int Mi, long double S ) :
    _tmsl( 0.0L ),
    _ptzi( 0 ),
    _readyflags( 0 ),
    _distortionflags( 0 )
{
    setgall( Y, Mo, D, H, Mi, S );
}

/**************************************************************************************************************/
gtimet::gtimet( long double Y, int Yd, int H, int Mi, long double S ) :
    _tmsl( 0.0L ),
    _ptzi( 0 ),
    _readyflags( 0 ),
    _distortionflags( 0 )
{
    setgall( Y, Yd, H, Mi, S );
}

/**************************************************************************************************************/
gtimet::gtimet( const gtimet& Gtime ) :
    _tmsl( Gtime._tmsl ),
    _ptzi( 0 ),
    _readyflags( 0 ),
    _distortionflags( 0 )
{
}

/**************************************************************************************************************/
gtimet::~gtimet()               
{
}

/**************************************************************************************************************/
gtimet&
gtimet::operator =( const gtimet& Rhs )
{
    _tmsl = Rhs._tmsl;
    _readyflags = 0;
    _distortionflags = 0;

    return *this;
}

/**************************************************************************************************************/
bool
gtimet::operator ==( const gtimet& Rhs ) const
{
    return _tmsl == Rhs._tmsl;
}

/**************************************************************************************************************/
bool
gtimet::operator <( const gtimet& Rhs ) const
{
    return _tmsl < Rhs._tmsl;
}

/**************************************************************************************************************/
string
gtimet::str() const
/*
    ISO-8601 representation
*/
{
    ostringstream  oss;

    if ( isabsolutebegin() )
    {
	oss << "-infinity";
    }
    else if ( isabsoluteend() )
    {
	oss << "infinity";
    }
    else
    {
	oss <<
	    fixed << setprecision(0) << gyear() <<
	    "-" << setfill('0') << setw(2) << gmonth()+1 <<
	    "-" << setfill('0') << setw(2) << gmday()+1 <<
	    " " << setfill('0') << setw(2) << ghour() <<
	    ":" << setfill('0') << setw(2) << gminute() <<
	    ":" << fixed << setprecision(0) << setfill('0') << setw(2) << floor(gsecond()) <<
	    "+0000";
    }

    return oss.str();
}

/**************************************************************************************************************/
bool
gtimet::isdistorted( fbitt FieldsMask ) const
{
    if 
    ( 
	//((FieldsMask & newflags & WDAYBIT) && (gwday() != newwday)) ||
	( FieldsMask & _distortionflags & SECONDBIT ) ||
	( FieldsMask & _distortionflags & MINUTEBIT ) ||
	( FieldsMask & _distortionflags & HOURBIT ) ||
	( FieldsMask & _distortionflags & MDAYBIT ) ||
	( FieldsMask & _distortionflags & MONTHBIT ) ||
	( FieldsMask & _distortionflags & YEARBIT ) ||
	( FieldsMask & _distortionflags & YDAYBIT ) ||
	( FieldsMask & _distortionflags & LSECONDBIT ) ||
        ( FieldsMask & _distortionflags & LMINUTEBIT ) ||
	( FieldsMask & _distortionflags & LHOURBIT ) ||
	( FieldsMask & _distortionflags & LMDAYBIT ) ||
	( FieldsMask & _distortionflags & LMONTHBIT ) ||
	( FieldsMask & _distortionflags & LYEARBIT )
    )
    {
	return true;
    }
    else
    {
	return false;
    }
}

/**************************************************************************************************************/
long double
gtimet::gsecond() const
{
    if ( _isready(SECONDBIT) )
    {
	presecond = _mod( _tms(), 60 );
	
	int  secn( static_cast<int>(presecond) );
	
	if ( secn == 59 )
	{
	    long  lscn( (long)floor(_tmsl/(MEANYEAR/2)+(55+HYEPSILON)) );		// leap second candidate number

	    if ( lscn < 0 )
	    {
		lscn = 0;
	    }
	    else if ( lscn >= LSTBLNUM )
	    {
		lscn = LSTBLNUM - 1;
	    }

	    const lsct*	 plsc( &LSTBL[lscn] );						// leap second candidate pointer
	    long double	 _tmsln( floor(_tmsl) );					// integer part of _tmsl

	    if ( plsc->lssecond==60 && _tmsln==plsc->lstmsl )
	    {
		presecond++;
	    }
	}
    }

    return presecond;
}

/**************************************************************************************************************/
int
gtimet::gminute() const
{
    if ( _isready(MINUTEBIT) )
    {
        preminute = static_cast<int>( _mod(_tms()/MEANMINUTE, 60) );
    }

    return preminute;
}

/**************************************************************************************************************/
int
gtimet::ghour() const
{
    if ( _isready(HOURBIT) )
    {
        prehour = static_cast<int>( _mod(_tms()/MEANHOUR, 24) );
    }

    return prehour;
}

/**************************************************************************************************************/
long double
gtimet::gdays() const
{
    if ( _isready(DAYSBIT) )
    {
        predays  = floor( _tms()/MEANDAY );
    }

    return predays;
}

/**************************************************************************************************************/
int
gtimet::gmday() const
{
    if ( _isready(MDAYBIT) )
    {
        premday  = gyday() - _tableyday( gyear(), gmonth() );
    }

    return premday;
}

/**************************************************************************************************************/
int
gtimet::gmonth() const
{
    if ( _isready(MONTHBIT) )
    {
        premonth = ( gyday() >> 5 ) + 1;

        if ( _tableyday( gyear(), premonth ) > gyday() )
        {
            premonth--;
        }
    }

    return premonth;
}

/**************************************************************************************************************/
long double
gtimet::gyear() const
{
    if ( _isready(YEARBIT) )
    {
	long double  days( gdays() );
	int	     dmn( static_cast<int>(_mod(days/(MEANYEAR/86400),400)) );
	
	preyear = floor( (gdays()+DAYSMAGICTBL[dmn])/(MEANYEAR/86400) ) + 2000;
    }
    
    return preyear;
}

/**************************************************************************************************************/
int
gtimet::gyday() const
{
    if ( _isready(YDAYBIT) )
    {
	long double  days( gdays() );
	int	     dmn( static_cast<int>(_mod(days/(MEANYEAR/86400),400)) );
	
        preyday  = static_cast<int>( _mod(gdays()+DAYSMAGICTBL[dmn], (MEANYEAR/86400)) );
    }
    
    return preyday;
}

/**************************************************************************************************************/
gtimet::wdayt
gtimet::gwday() const
{
    if ( _isready(WDAYBIT) )
    {
        prewday  = wdayt( static_cast<int>(_mod(gdays()+BASEWEEKDAY,7)) );
    }
    
    return prewday;
}

/**************************************************************************************************************/
long double
gtimet::juliandate() const
{
    return (_tmsl+TMSLTTDELTA)/MEANDAY + BASEJULIANDATE;
}

/**************************************************************************************************************/
jdnt
gtimet::juliandaynumber() const
{
    long double	 jd = juliandate();
    
    assert( jd > numeric_limits<jdnt>::min() && jd < numeric_limits<jdnt>::max() );
    
    return static_cast<jdnt>( jd );
}

/**************************************************************************************************************/
long double
gtimet::modifiedjuliandate() const
{
    return (_tmsl+TMSLTTDELTA)/MEANDAY + BASEMODIFIEDJULIANDATE;
}

/**************************************************************************************************************/
string
gtimet::tz()
{
    const char*	 pctz( getenv("TZ") );

    return pctz? pctz : string();
}

/**************************************************************************************************************/
time_t
gtimet::unixtime( lsmodet Lsmode ) const
{
    long double	 tmc;

    if ( Lsmode == LSON || (Lsmode == LSTZ && islstimezone()) )
    {
	tmc = _tmsl - TMSL19700101;
    }
    else
    {
	tmc = _tms() - TMS19700101;
    }

    if ( tmc < numeric_limits<time_t>::min() )
    {
	tmc = numeric_limits<time_t>::min();
    }

    if ( tmc > numeric_limits<time_t>::max() )
    {
	tmc = numeric_limits<time_t>::max();
    }

    return static_cast<time_t>( tmc );
}

/**************************************************************************************************************/
long double
gtimet::durationfrom( const gtimet& Begin ) const
{
    return _tmsl - Begin._tmsl;
}

/**************************************************************************************************************/
long double
gtimet::_tms() const
{
    if ( _isready(TMSBIT) )
    {
	long  lscn( (long)floor(_tmsl/(MEANYEAR/2)+(55+HYEPSILON)) );		// leap second candidate number

	if ( lscn < 0 )
	{
	    lscn = 0;
	}
	else if ( lscn >= LSTBLNUM )
	{
	    lscn = LSTBLNUM - 1;
	}

	const lsct*  plsc( &LSTBL[lscn] );					// leap second candidate pointer
	long double  tmsln( floor(_tmsl) );					// integer part of _tmsl

	pretms = _tmsl + plsc->lstms - plsc->lstmsl;

	if 
	(
	    ( tmsln>plsc->lstmsl && plsc->lssecond==60 ) ||
	    ( tmsln==plsc->lstmsl && plsc->lssecond==60 )
	)
	{
	    pretms--;
	}
	else if 
	(
	    plsc->lstmsl==tmsln && plsc->lssecond==58
	)
	{
	    pretms+=2;
	}
    }

    return pretms;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setgsecond( long double Second )
{
    // What if _tmsl loss Second precision ?
    _settms( _tms() - gsecond() );
    _tmsl += Second;
    
    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setgminute( int Minute )
{
    _settms( _tms() - (gminute()-Minute)*MEANMINUTE );
    
    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setghour( int Hour )
{
    _settms( _tms() - (ghour()-Hour)*MEANHOUR );
    
    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setgmday( int Mday )
{
   _settms( _tms() - (gmday()-Mday)*MEANDAY );
    
    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setgmonth( int Month )
{
    int        	 oldday( gmday() );
    int        	 oldhour( ghour() );
    int        	 oldminute( gminute() );
    long double  oldsecond( gsecond() );
    
    _settms( _tms() - (gmonth()-Month)*(MEANYEAR/12) );
    _arrangeday( oldday );
    _setgmday( oldday );
    _setghour( oldhour );
    _setgminute( oldminute );
    _setgsecond( oldsecond );
    
    if ( gmday() != oldday )
    {
	_setgmday( -1 );
    }

    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setgyear( long double Year )
{
    int 	 oldmonth( gmonth() );
    int  	 oldday( gmday() );
    int    	 oldhour( ghour() );
    int		 oldminute( gminute() );
    long double	 oldsecond( gsecond() );
    
    _settms( _tms() - (gyear()-Year)*MEANYEAR );
    _arrangeday( oldday );
    _setgmonth( oldmonth );
    _setgmday( oldday );
    _setghour( oldhour );
    _setgminute( oldminute );
    _setgsecond( oldsecond );

    if ( gmday() != oldday )
    {
	_setgmday( -1 );
    }
	
    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setgyday( int Yday )
{
    _settms( _tms() - (gyday()-Yday)*MEANDAY );

    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setgwday( wdayt Wday )
{
    _settms( _tms() - (gwday()-Wday)*MEANDAY );

    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_setgdays( long double Days )
{
    _settms( floor(Days)*MEANDAY + _mod(_tms(),MEANDAY) );

    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::_settms( long double Tms )
{
    long  lscn( (long)floor(Tms/(MEANYEAR/2)+(55+HYEPSILON)) );		// leap second candidate number

    if ( lscn < 0 )
    {
	lscn = 0;
    }
    else if ( lscn >= LSTBLNUM )
    {
	lscn = LSTBLNUM - 1;
    }
    
    const lsct*	 plsc( &LSTBL[lscn] );					// leap second candidate pointer
    long double	 tmsn( floor(Tms) );					// integer part of tms
    
    _tmsl = Tms + plsc->lstmsl - plsc->lstms;
    
    if 
    ( 
	( tmsn > plsc->lstms && plsc->lssecond == 60 ) ||
	( tmsn == plsc->lstms && plsc->lssecond == 60 )
    )
    {
	_tmsl++;
    }
    
    // pretms = Tms;
    // _readyflags = TMSBIT;
    _readyflags = 0; 		// TMS of 23:59:59 for LS58 should be distorted
    
    return *this;
}

/**************************************************************************************************************/
void
gtimet::setabsolutebegin()
{
    _tmsl = std::numeric_limits<long double>::min();
    _readyflags = 0;
    _distortionflags = 0;
}

/**************************************************************************************************************/
void
gtimet::setabsoluteend()
{
    _tmsl = std::numeric_limits<long double>::max();
    _readyflags = 0;
    _distortionflags = 0;
}

/**************************************************************************************************************/
void
gtimet::setunixtime( time_t Sec1970, lsmodet Lsmode )
{
    _readyflags = 0;
    _distortionflags = 0;

    if ( Lsmode == LSON || (Lsmode == LSTZ && islstimezone()) )
    {    
	_tmsl = TMSL19700101 + Sec1970;
    }
    else
    {
	_settms( TMS19700101 + Sec1970 );
    }
}

/**************************************************************************************************************/
gtimet&
gtimet::setjuliandate( long double Jd )
{
    _tmsl = (Jd - BASEJULIANDATE) * MEANDAY - TMSLTTDELTA;
    _readyflags = 0;
    
    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::setjuliandaynumber( jdnt Jdn )
{
    _tmsl = (Jdn - BASEJULIANDATE) * MEANDAY - TMSLTTDELTA + _mod(_tmsl, MEANDAY);
    _readyflags = 0;
    
    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::setmodifiedjuliandate( long double Mjd )
{
    _tmsl = (Mjd - BASEMODIFIEDJULIANDATE) * MEANDAY - TMSLTTDELTA;
    _readyflags = 0;
    
    return *this;
}

/**************************************************************************************************************/
bool
gtimet::settz( const string& Tz )
{
    string::size_type  ntzpos( Tz.find_first_not_of(" \t") );
    string::size_type  ntzsize( (ntzpos >= Tz.size())? 0 : Tz.find_last_not_of(" \t") - ntzpos + 1 );
    string	       ntz;

    if ( ntzsize > 0 )
    {
	ntz = Tz.substr( ntzpos, ntzsize );
    }

    if ( ntz.size() == 0 )
    {
	unsetenv( "TZ" );
    }
    else
    {
	string  tzs;

	if ( ntz[0] != ':' )
	{
	    tzs = ntz;
	}
	else
	{
	    tzs = ":";

            if ( ntz.size() > 1 )
            {
                string    fname( ntz.substr(1,string::npos) );
                ifstream  ifs( fname.c_str() );

		if ( ntz[1] == '/' && !ifs )
		{
		    return false;
		}
                else if ( ntz[1] != '/' && !!ifs )
                {
                    bool  istzfile( true );

#if !defined( __MACH__ )

                    char  tzmagic[5];

                    ifs.get( tzmagic, sizeof(tzmagic) );

                    if ( string(tzmagic) != "TZif" )
                    {
                        istzfile = false;
                    }
                    
#endif

                    if ( istzfile )
                    {
                        char*  pwd( new char[PATH_MAX+1] );

                        if ( !getcwd(pwd,PATH_MAX) )
                        {
                            delete[] pwd;

                            return false;
                        }

                        tzs += pwd;
                        tzs += "/";

                        delete[] pwd;
                    }
                }

                tzs += fname;
            }
	}
    
	if ( setenv("TZ", tzs.c_str(), 1) )
	{
	    return false;
	}
    }

    tzset();

    return true;
}

/**************************************************************************************************************/
gtimet&
gtimet::setgall( long double Y, int Mo, int D, int H, int Mi, long double S )
{
    _setgyear( Y );
    _setgmonth( Mo );
    _setgmday( D );
    _setghour( H );
    _setgminute( Mi );
    _setgsecond( S );

    _distortionflags = 0;

    if ( gyear() != Y )
    {
	_distortionflags |= YEARBIT;
    }

    if ( gmonth() != Mo )
    {
	_distortionflags |= MONTHBIT;
    }

    if ( gmday() != D )
    {
	_distortionflags |= MDAYBIT;
    }

    if ( ghour() != H )
    {
	_distortionflags |= HOURBIT;
    }

    if ( gminute() != Mi )
    {
	_distortionflags |= MINUTEBIT;
    }

    if ( gsecond() != S )
    {
	_distortionflags |= SECONDBIT;
    }

    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::setgall( long double Y, int Yd, int H, int Mi, long double S )
{
    _setgyear( Y );
    _setgyday( Yd );
    _setghour( H );
    _setgminute( Mi );
    _setgsecond( S );

    _distortionflags = 0;

    if ( gyear() != Y )
    {
	_distortionflags |= YEARBIT;
    }

    if ( gyday() != Yd )
    {
	_distortionflags |= YDAYBIT;
    }

    if ( ghour() != H )
    {
	_distortionflags |= HOURBIT;
    }

    if ( gminute() != Mi )
    {
	_distortionflags |= MINUTEBIT;
    }

    if ( gsecond() != S )
    {
	_distortionflags |= SECONDBIT;
    }

    return *this;
}

/**************************************************************************************************************/
gtimet&
gtimet::setall( const string& Tmstr, string* Perrdesc )
/*
    Recognize ISO-8601 Extended format:
        [+/-Y*]YYYY-MM-DD		calendar date [expanded] complete representation
        [+/-Y*]YYYY-DDD			ordinal date [expanded] complete representation
        [+/-Y*]YYYY-Www[-D]		week date [expanded] representation
	[T/ *]hh:mm:ss[(,/.)s*]		time of the day complete representation [with decimal fraction]
    	[T/ *]hh:mm[(,/.)m*]		time of the day reduced precision representation [with decimal fraction]
	[ *](Z/(+/-)hh[:mm])		time zone representation

    Recognize ISO-8601 Basic format:
	[(+/-)Y*]YYYYMMDD		calendar date [expanded] complete representation
        [+/-Y*]YYYYWww[D]		week date [expanded] representation
	[T/ *]hhmmss[(,/.)s*]		time of the day complete representation [with decimal fraction]
	[ *](Z/(+/-)hh[mm])		time zone representation
	
    Recognize PostgreSQL special time values:
	-infinity
	infinity

    Recognize special time values:
	absolute_begin, absolutebegin
	absolute_end, absoluteend

    To avoid ambiguity do not recognize:
	all truncated representations
	basic format ordinal date
	basic format reduced precision representations
*/
/*
    TODO
	1) Правильное выставление флагов distortion
	2) Создать методы gweek(), setgweek()
	3) Разработать доскональные тесты
	4) Попытаться оптимизировать алгоритм использованием сигнатур
*/
{
    string::size_type  ntmstrpos( Tmstr.find_first_not_of(" \t") );
    string::size_type  ntmstrsize( (ntmstrpos >= Tmstr.size())? 0 : Tmstr.find_last_not_of(" \t") - ntmstrpos + 1 );
    string	       ntmstr;

    if ( ntmstrsize > 0 )
    {
	ntmstr = Tmstr.substr( ntmstrpos, ntmstrsize );
    }

    if ( ntmstr == "-infinity" || ntmstr == "absolute_begin" || ntmstr == "absolutebegin" )
    {
	setabsolutebegin();

	return *this;
    }
    else if ( ntmstr == "infinity" || ntmstr == "absolute_end" || ntmstr == "absoluteend" )
    {
	setabsoluteend();

	return *this;
    }

    const char*	 faildesc( NULL );
    bool	 basicformatflag( ntmstr.find_first_of("-:") == string::npos );
    string	 datepart;
    string	 timepart;
    string	 zonepart;
    long double	 year( numeric_limits<long long>::min() );
    int		 month( -1 );
    int		 mday( -1 );
    int		 week( -1 );
    int		 wday( -1 );
    int		 yday( -1 );
    int		 hour( -1 );
    double	 minute( -1.0 );
    double	 second( -1.0 );
    char	 tzsign( '+' );
    int		 tzhour( -1 );
    int		 tzminute( -1 );

    string::size_type  partb( ntmstr.find_first_not_of(' ') );
    string::size_type  parte( ntmstr.find_first_of("T ",partb) );

    if ( partb == string::npos )
    {
    	faildesc = "empty string";
    }
    else
    {
	string  partstr( ntmstr, partb, parte-partb );

	if ( partstr.find_first_not_of("+-W0123456789") == string::npos && partstr.size() > 7 )
    	{
	    //**** date part *********************************************************************************

	    char  c0l( partstr[0] );
	    char  c2r( partstr[partstr.size()-2] );
	    char  c3r( partstr[partstr.size()-3] );
	    char  c4r( partstr[partstr.size()-4] );
	    char  c5r( partstr[partstr.size()-5] );
	    char  c6r( partstr[partstr.size()-6] );

	    if ( !isdigit(c0l) && c0l != '-' && c0l != '+' )
	    {
	    	faildesc = "incorrect first symbol in the date part";
	    }
	    else if ( c3r == '-' && c6r == '-' )
	    {
	        // extended calendar date

	        string	yearstr( partstr, 0, partstr.size()-6 );
	        string	monthstr( partstr, partstr.size()-5, 2 );
	        string	mdaystr( partstr, partstr.size()-2 );

	        if ( yearstr.find_first_not_of("0123456789",1) != string::npos )
	        {
		    faildesc = "incorrect symbol in the extended format year part";
	    	}
	    	else if ( monthstr.find_first_not_of("0123456789") != string::npos )
	    	{
		    faildesc = "incorrect symbol in the extended format month part";
	    	}
	    	else if ( mdaystr.find_first_not_of("0123456789") != string::npos )
	    	{
		    faildesc = "incorrect symbol in the extended format month day part";
	    	}
		else
		{
		    datepart = partstr;
	    	    year = atoll( yearstr.c_str() );
	    	    month = atoi( monthstr.c_str() );
	    	    mday = atoi( mdaystr.c_str() );
		}
	    }
	    else if ( c4r == '-' )
	    {
	    	// extended ordinal date or extended week date

	    	string	yearstr( partstr, 0, partstr.size()-4 );

	    	if ( yearstr.find_first_not_of("0123456789",1) != string::npos )
	    	{
		    faildesc = "incorrect symbol in the extended format year part";
	    	}
	    	else if ( c3r == 'W' )
	    	{
		    // extended week date

		    string  weekstr( partstr, partstr.size()-2 );

	            if ( weekstr.find_first_not_of("0123456789") != string::npos )
		    {
		    	faildesc = "incorrect symbol in the extended format week part";
		    }
		    else
		    {
			datepart = partstr;
	    	    	year = atoll( yearstr.c_str() );
	            	week = atoi( weekstr.c_str() );
		    }
	    	}
	    	else
	    	{
		    // extended ordinal date

		    string  ydaystr( partstr, partstr.size()-3 );

	            if ( ydaystr.find_first_not_of("0123456789") != string::npos )
		    {
		    	faildesc = "incorrect symbol in the extended format day of year part";
		    }
		    else
		    {
			datepart = partstr;
	    	    	year = atoll( yearstr.c_str() );
	            	yday = atoi( ydaystr.c_str() );
		    }
	    	}
	    }
	    else if ( c2r == '-' && c5r == 'W' && c6r == '-' )
	    {
	    	// extended week date with wday

	    	string	yearstr( partstr, 0, partstr.size()-6 );
	    	string	weekstr( partstr, partstr.size()-4, 2 );
	    	string	wdaystr( partstr, partstr.size()-1 );

	    	if ( yearstr.find_first_not_of("0123456789",1) != string::npos )
	    	{
		    faildesc = "incorrect symbol in the extended format year part";
	    	}
	    	else if ( weekstr.find_first_not_of("0123456789") != string::npos )
	    	{
		    faildesc = "incorrect symbol in the extended format week part";
	    	}
	    	else if ( wdaystr.find_first_not_of("0123456789") != string::npos )
	    	{
		    faildesc = "incorrect symbol in the extended format day of week part";
	    	}
		else
		{
		    datepart = partstr;
	    	    year = atoll( yearstr.c_str() );
	    	    week = atoi( weekstr.c_str() );
	    	    wday = atoi( wdaystr.c_str() );
		}
	    }
	    else if ( c3r == 'W' )
	    {
	    	// basic week date

	    	string	yearstr( partstr, 0, partstr.size()-3 );
	    	string	weekstr( partstr, partstr.size()-2 );

	    	if ( yearstr.find_first_not_of("0123456789",1) != string::npos )
	    	{
		    faildesc = "incorrect symbol in the basic format year part";
	    	}
	    	else if ( weekstr.find_first_not_of("0123456789") != string::npos )
	    	{
		    faildesc = "incorrect symbol in the basic format week part";
	    	}
		else
		{
		    datepart = partstr;
	    	    year = atoll( yearstr.c_str() );
	    	    week = atoi( weekstr.c_str() );
		}
	    }
	    else if ( c4r == 'W' )
	    {
	    	// basic week date with wday

	    	string	yearstr( partstr, 0, partstr.size()-4 );
	    	string	weekstr( partstr, partstr.size()-3, 2 );
	    	string	wdaystr( partstr, partstr.size()-1 );

	    	if ( yearstr.find_first_not_of("0123456789",1) != string::npos )
	    	{
		    faildesc = "incorrect symbol in the basic format year part";
	    	}
	    	else if ( weekstr.find_first_not_of("0123456789") != string::npos )
	    	{
		    faildesc = "incorrect symbol in the basic format week part";
	    	}
	    	else if ( wdaystr.find_first_not_of("0123456789") != string::npos )
	    	{
		    faildesc = "incorrect symbol in the basic format day of week part";
		}
		else
		{
		    datepart = partstr;
	    	    year = atoll( yearstr.c_str() );
	    	    week = atoi( weekstr.c_str() );
	    	    wday = atoi( wdaystr.c_str() );
		}
	    }
	    else if ( partstr.find_first_not_of("0123456789",1) == string::npos )
	    {
	    	// basic calendar date

	    	string	yearstr( partstr, 0, partstr.size()-4 );
	    	string	monthstr( partstr, partstr.size()-4, 2 );
	    	string	mdaystr( partstr, partstr.size()-2 );

		datepart = partstr;
	    	year = atoll( yearstr.c_str() );
	    	month = atoi( monthstr.c_str() );
	    	mday = atoi( mdaystr.c_str() );
	    }
	}

	if ( year > numeric_limits<long long>::min() )
	{
	    partb = ntmstr.find_first_not_of( "T ", parte );
	}
	else
	{
	    partb = ntmstr.find_first_not_of( "T ", partb );
	}

	if ( partb == string::npos )
	{
	    partstr = "";
	}
	else
	{
	    parte = ntmstr.find_first_of( " Z-+", partb );
	    partstr = ntmstr.substr( partb, parte-partb );
	}

	if ( partstr.find_first_not_of(":.,0123456789") == string::npos && partstr.size() > 0 )
    	{
	    //**** time part *********************************************************************************

	    if ( basicformatflag )
    	    {
	    	// basic time

	    	string		   hourstr( partstr, 0, 2 );
	    	string		   minutestr( partstr, 2, 2 );
	    	string		   secondstr( partstr, 4 );
	    	string::size_type  commai( secondstr.find_first_of(',') );

	    	if ( commai != string::npos )
	    	{
	    	    secondstr[ commai ] = '.';
	    	}

	    	if ( hourstr.find_first_not_of("0123456789") != string::npos )
	    	{
	    	    faildesc = "incorrect symbol in the basic format hour part";
	    	}
	    	else if ( minutestr.find_first_not_of("0123456789") != string::npos )
	    	{
	    	    faildesc = "incorrect symbol in the basic format minute part";
	    	}
	    	else if ( secondstr.find_first_not_of(".0123456789") != string::npos )
	    	{
	    	    faildesc = "incorrect symbol in the basic format second part";
	    	}
	    	else if ( !isdigit(secondstr[0]) || !isdigit(secondstr[1]) )
	    	{
	    	    faildesc = "the basic format second's integer part less than 2 symbol";
	    	}
	    	else if ( secondstr.size() > 2 && secondstr[2] != '.' )
	    	{
	    	    faildesc = "the extended format second's integer part more than 2 symbol";
	    	}
		else
		{
	    	    char*   eptr;
		    double  presecond( strtod(secondstr.c_str(),&eptr) );

	    	    if ( eptr-secondstr.c_str() < secondstr.size() )
	    	    {
	    	    	faildesc = "more than one symbol '.' or ',' in the basic format second part";
	    	    }
		    else
		    {
			timepart = partstr;
	    	    	hour = atoi( hourstr.c_str() );
	    	    	minute = strtod( minutestr.c_str(), NULL );
	    	    	second = presecond;
		    }
		}
	    }
    	    else if ( partstr.size() > 5 && partstr[5] == ':' )
    	    {
	    	// extended complete time

	    	string		   hourstr( partstr, 0, 2 );
	    	string		   minutestr( partstr, 3, 2 );
	    	string		   secondstr( partstr, 6 );
	    	string::size_type  commai( secondstr.find_first_of(',') );

	    	if ( commai != string::npos )
	    	{
	    	    secondstr[ commai ] = '.';
	    	}

	    	if ( hourstr.find_first_not_of("0123456789") != string::npos )
	    	{
	    	    faildesc = "incorrect symbol in the extended format hour part";
	    	}
	    	else if ( minutestr.find_first_not_of("0123456789") != string::npos )
	    	{
	    	    faildesc = "incorrect symbol in the extended format minute part";
	    	}
	    	else if ( secondstr.find_first_not_of(".0123456789") != string::npos )
	    	{
	    	    faildesc = "incorrect symbol in the extended format second part";
	    	}
	    	else if ( !isdigit(secondstr[0]) || !isdigit(secondstr[1]) )
	    	{
	    	    faildesc = "the extended format second's integer part less than 2 symbol";
	    	}
	    	else if ( secondstr.size() > 2 && secondstr[2] != '.' )
	    	{
	    	    faildesc = "the extended format second's integer part more than 2 symbol";
	    	}
		else
		{
	    	    char*   eptr;
		    double  presecond( strtod(secondstr.c_str(),&eptr) );

	    	    if ( eptr-secondstr.c_str() < secondstr.size() )
	    	    {
	    	    	faildesc = "more than one symbol '.' or ',' in the basic format second part";
	    	    }
		    else
		    {
			timepart = partstr;
	    	    	hour = atoi( hourstr.c_str() );
	    	    	minute = strtod( minutestr.c_str(), NULL );
	    	    	second = presecond;
		    }
		}
	    }
    	    else if ( partstr.size() < 5 || partstr[2] != ':' )
    	    {
	    	faildesc = "incorrect extended format hour part";
	    }
	    else
	    {
	    	// extended time with reduced precision

	    	string		   hourstr( partstr, 0, 2 );
	    	string		   minutestr( partstr, 3 );
	    	string::size_type  commai( minutestr.find_first_of(',') );

	    	if ( commai != string::npos )
	    	{
	    	    minutestr[ commai ] = '.';
	    	}

	    	if ( hourstr.find_first_not_of("0123456789") != string::npos )
	    	{
	    	    faildesc = "incorrect symbol in the extended format hour part";
	    	}
	    	else if ( minutestr.find_first_not_of(".0123456789") != string::npos )
	    	{
	    	    faildesc = "incorrect symbol in the extended format minute part";
	    	}
	    	else if ( !isdigit(minutestr[0]) || !isdigit(minutestr[1]) )
	    	{
	    	    faildesc = "the extended format minute's integer part less than 2 symbol";
	    	}
	    	else if ( minutestr.size() > 2 && minutestr[2] != '.' )
	    	{
	    	    faildesc = "the extended format minute's integer part more than 2 symbol";
	    	}
		else
		{
	    	    char*   eptr;
		    double  preminute( strtod(minutestr.c_str(),&eptr) );

	    	    if ( eptr-minutestr.c_str() < minutestr.size() )
	    	    {
	    	    	faildesc = "more than one symbol '.' or ',' in the extended format minute part";
	    	    }
		    else
		    {
			timepart = partstr;
	    	    	hour = atoi( hourstr.c_str() );
	    	    	minute = preminute;
		    }
		}
    	    }
	}

	if ( parte != string::npos )
    	{
	    string::size_type  partb( ntmstr.find_first_not_of(' ',parte) );

	    if ( partb != string::npos )
	    {
	    	string::size_type  parte( ntmstr.find_last_not_of(' ') + 1 );

		partstr = ntmstr.substr( partb, parte-partb );

		if ( partstr.find_first_not_of("Z-+:0123456789") != string::npos )
	    	{
		    faildesc = "incorrect symbol in the zone part";
	    	}

	        //**** zone part *****************************************************************************

		char  c0l( partstr[0] );

		if ( c0l == 'Z' )
		{
		    if ( partstr.size() > 1 )
		    {
			faildesc = "too large zone part";
		    }
		    else
		    {
			zonepart = partstr;
		    	tzhour = 0;
		    	tzminute = 0;
		    }
		}
		else if ( c0l != '-' && c0l != '+' )
		{
		    faildesc = "incorrect symbol as the sign of time zone";
		}
		else
		{
		    if ( c0l == '-' )
		    {
		    	tzsign = '-';
		    }

		    if ( partstr.size() == 3 )
		    {
			// time zone with reduced precision

			string  tzhourstr( partstr, 1, 2 );

			if ( tzhourstr.find_first_not_of("0123456789") != string::npos )
			{
			    faildesc = "incorrect symbol in the time zone hour part";
			}
			else
			{
			    zonepart = partstr;
			    tzhour = atoi( tzhourstr.c_str() );
			    tzminute = 0;
			}
		    }
		    else if ( partstr[3] == ':' )
		    {
			// extended time zone

			string  tzhourstr( partstr, 1, 2 );
			string	tzminutestr( partstr, 4, 2 );

			if ( partstr.size() != 6 )
			{
			    faildesc = "incorrect size of the extended format zone part";
			}
			else if ( tzhourstr.find_first_not_of("0123456789") != string::npos )
			{
			    faildesc = "incorrect symbol in the extended format time zone hour part";
			}
			else if ( tzminutestr.find_first_not_of("0123456789") != string::npos )
			{
			    faildesc = "incorrect symbol in the extended format time zone minute part";
			}
			else
			{
			    zonepart = partstr;
			    tzhour = atoi( tzhourstr.c_str() );
			    tzminute = atoi( tzminutestr.c_str() );
			}
		    }
		    else
		    {
			// basic time zone

			string	tzhourstr( partstr, 1, 2 );
			string	tzminutestr( partstr, 3, 2 );

			if ( partstr.size() != 5 )
			{
			    faildesc = "incorrect size of the basic format zone part";
			}
	 		else if ( tzhourstr.find_first_not_of("0123456789") != string::npos )
			{
			    faildesc = "incorrect symbol in the basic format time zone hour part";
			}
			else if ( tzminutestr.find_first_not_of("0123456789") != string::npos )
			{
			    faildesc = "incorrect symbol in the basic format time zone minute part";
			}
			else
			{
			    zonepart = partstr;
			    tzhour = atoi( tzhourstr.c_str() );
			    tzminute = atoi( tzminutestr.c_str() );
			}
		    }
		}
	    }
    	}
    }

    double  ssecond( ((second>=0)?second:0) + ((minute>=0)?minute*MEANMINUTE:0) );
    ltimet  now( time(NULL) );
    gtimet  gtm( now.unixtime() );

    if ( tzhour >= 0 )
    {                   
        gtm.setgall( now.gyear(), now.gmonth(), now.gmday(), 0, 0, 0  );
    }                   
    else                    
    {                   
        gtm.setgall( now.lyear(), now.lmonth(), now.lmday(), 0, 0, 0  );
    }                   

    if ( year > numeric_limits<long long>::min() )
    {
	gtm._setgyear( year );
    }

    if ( month >= 0 )
    {
	gtm._setgmonth( month - 1 );

	if ( !faildesc && month < 1 && month > 12 )
	{
	     faildesc = "month is out of range";
	}
    }

    if ( mday >= 0 )
    {
	gtm._setgmday( mday - 1 );

	if ( !faildesc && mday-1 != gtm.gmday() )
	{
	     faildesc = "day of the month is out of range";
	}
    }

    //if ( week >= 0 )		/**/;

    if ( wday >= 0 )
    {
	gtm._setgwday( static_cast<wdayt>(wday-1) );

	if ( !faildesc && wday < 1 && wday > 7 )
	{
	     faildesc = "day of the week is out of range";
	}
    }

    if ( yday >= 0 )
    {
	gtm._setgyday( yday - 1 );

	if ( !faildesc && yday-1 != gtm.gyday() )
	{
	     faildesc = "day of the year is out of range";
	}
    }

    if ( hour >= 0 )
    {
	gtm._setghour( hour );

	if ( !faildesc && hour > 23 )
	{
	     faildesc = "day of the year is out of range";
	}
    }

    gtm._setgsecond( ssecond );

    if ( !faildesc && minute >= 60 )
    {
	faildesc = "minute is out of range";
    }

    if ( second >= 0 )
    {	
    	if ( !faildesc && second != gtm.gsecond() )		// may be leap second
    	{
	    faildesc = "second is out of range";
    	}
    }

    if ( tzhour >= 0 )
    {
	gtm -= tzhour * static_cast<int>(MEANHOUR) + tzminute * static_cast<int>(MEANMINUTE);
	operator =( gtm );
    }
    else
    {
	ltimet	ltm( gtm.gyear(), gtm.gmonth(), gtm.gmday(), gtm.ghour(), gtm.gminute(), gtm.gsecond() );

	operator =( ltm );
    }

    if ( faildesc && Perrdesc )
    {
	(*Perrdesc = faildesc) += ": "
	    "datepart=\"" + datepart + "\"; "
	    "timepart=\"" + timepart + "\"; "
	    "zonepart=\"" + zonepart + "\";";
    }

    return *this;
}

/**************************************************************************************************************/
bool
gtimet::islstimezone()
{
    struct tm  tmdata;
    time_t     tm( 78796801 );		// 1972-07-01 00:00:00Z

#if !defined( __linux__ )

    tm = time2posix( tm );

#endif

    gmtime_r( &tm, &tmdata );

    return tmdata.tm_sec == 0;
}

/**************************************************************************************************************/
const char*
gtimet::monthfull( int Month )
{
    assert( Month >=0 && Month < 12 );

    const char*  monthstbl[] =
    {
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December"
    };

    return monthstbl[ Month ];
}

/**************************************************************************************************************/
const char*
gtimet::monthabbr( int Month )
{
    assert( Month >=0 && Month < 12 );

    const char*  amonthstbl[] =
    {
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec"
    };

    return amonthstbl[ Month ];
}

/**************************************************************************************************************/
const char*
gtimet::wdayfull( gtimet::wdayt Wday )
{
    assert( Wday >=0 && Wday < 7 );

    const char*  wdaystbl[] =
    {
        "Sunday",
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday"
    };

    return wdaystbl[ Wday ];
}

/**************************************************************************************************************/
const char*
gtimet::wdayabbr( gtimet::wdayt Wday )
{
    assert( Wday >=0 && Wday < 7 );

    const char*  awdaystbl[] =
    {
        "Sun",
        "Mon",
        "Tue",
        "Wed",
        "Thu",
        "Fri",
        "Sat"
    };

    return awdaystbl[ Wday ];
}

/**************************************************************************************************************/
const char*
gtimet::gampm() const
{
    return ( ghour() > 11 )? "pm" : "am";
}

/**************************************************************************************************************/
int
gtimet::ghour112() const
{
    return ( ghour() + 11 ) % 12 + 1;
}

/**************************************************************************************************************/
gtimet&
gtimet::rand( long double MeanInterTime )
{
    return (*this) -= MeanInterTime * log( _Drand1() );
}

/**************************************************************************************************************/
int
gtimet::_tableyday( long double Y, int M )
{
    static int  DaysOfYear[]     = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 };
    static int  DaysOfLeapYear[] = { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 };

    return ( (_mod(Y,4)||(_mod(Y,100)==0&&_mod(Y,400)))? DaysOfYear : DaysOfLeapYear )[ M ];
}

/**************************************************************************************************************/
long double
gtimet::_mod( long double Dividend, long double Divisor )
{
    double       quotient;
    long double  product;

    if ( Divisor == 0 || modf(Dividend/Divisor,&quotient) == 0 )
    {
	return 0;
    }
    
    product = floor(Dividend/Divisor) * Divisor;

    return Dividend - product;
}

/**************************************************************************************************************/
bool
gtimet::_isready( int Mask ) const
{
    if ( _readyflags & Mask )
    {
        return false;
    }
    else
    {
        _readyflags |= Mask;

        return true;
    }
}

/**************************************************************************************************************/
int
gtimet::gmdays() const
{
    static int   daysinmonthofyear[]  = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
    static int   daysinmonthoflyear[] = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
    long double	 yr = gyear();

    return ( (_mod(yr,4)||(_mod(yr,100)==0&&_mod(yr,400)))? daysinmonthofyear : daysinmonthoflyear )[ gmonth() ];
}

/**************************************************************************************************************/
void
gtimet::_arrangeday( int OldDay )
{
    const int  dayepsilon( 4 );
    int	       newday( gmday() );
    
    if ( newday > OldDay )
    {
        if ( newday - OldDay > dayepsilon )
        {
    	    _settms( _tms() + dayepsilon*MEANDAY );
	}
    }
    else if ( newday < OldDay )
    {
	if ( newday - OldDay < -dayepsilon )
	{
	    _settms( _tms() - dayepsilon*MEANDAY );
	}
    }
}

/**************************************************************************************************************/
ltimet::ltimet( long double Y, int Mo, int D, int H, int Mi, long double S ) :
    gtimet( 0 )
{
    setlall( Y, Mo, D, H, Mi, S );
}

/**************************************************************************************************************/
ltimet::ltimet( const string& TimeString, string* Perrdesc ) :
    gtimet( TimeString, Perrdesc )
{
}

/**************************************************************************************************************/
ltimet::~ltimet()
{
}

/**************************************************************************************************************/
long double
gtimet::lsecond() const
{
    return gsecond();
}

/**************************************************************************************************************/
int
gtimet::lminute() const
{
    return gtimet(*this)._settms(_tms()+lutcoffset()).gminute();
}

/**************************************************************************************************************/
int
gtimet::lhour() const
{
    return gtimet(*this)._settms(_tms()+lutcoffset()).ghour();
}

/**************************************************************************************************************/
int
gtimet::lmday() const
{
    return gtimet(*this)._settms(_tms()+lutcoffset()).gmday();
}

/**************************************************************************************************************/
int
gtimet::lmonth() const
{
    return gtimet(*this)._settms(_tms()+lutcoffset()).gmonth();
}

/**************************************************************************************************************/
long double
gtimet::lyear() const
{
    return gtimet(*this)._settms(_tms()+lutcoffset()).gyear();
}

/**************************************************************************************************************/
int
gtimet::lyday() const
{
    return gtimet(*this)._settms(_tms()+lutcoffset()).gyday();
}

/**************************************************************************************************************/
gtimet::wdayt
gtimet::lwday() const
{
    return gtimet(*this)._settms(_tms()+lutcoffset()).gwday();
}

/**************************************************************************************************************/
string
gtimet::ltzname() const
{
    /*if ( lisdst() > 0 )
    {
	return tzname[1];
    }
    else if ( lisdst() == 0 )
    {
	return tzname[0];
    }
    else
    {
	return "unknown time zone";
    }*/
    
    struct tm  locstm;
    time_t     tm( unixtime() );

    if ( tm < 0 )
    {
	gtimet	     gtm( *this );
	long double  year( gtm.gyear() );

	if ( _tmsl < TMSL19700101 )
	{
	    gtm._setgyear( (_mod(year,4)||(_mod(year,100)==0&&_mod(year,400)))? 1971 : 1972 );
	}
	else
	{
	    gtm._setgyear( (_mod(year,4)||(_mod(year,100)==0&&_mod(year,400)))? 1999 : 2000 );
	}

	tm = gtm.unixtime();
    }

    localtime_r( &tm, &locstm );

    return locstm.tm_zone;
}

/**************************************************************************************************************/
int
gtimet::ltzhour() const
{
    return lutcoffset() / 3600;
}

/**************************************************************************************************************/
int
gtimet::ltzminute() const
{
    int  tzmin( (lutcoffset()/60) % 60 );
    
    if ( tzmin < 0 )
    {
	tzmin = -tzmin;
    }
    
    return tzmin;
}

/**************************************************************************************************************/
int
gtimet::lisdst() const
{
    struct tm  locstm;
    time_t     tm( unixtime() );

    if ( tm < 0 )
    {
	gtimet	     gtm( *this );
	long double  year( gtm.gyear() );

	if ( _tmsl < TMSL19700101 )
	{
	    gtm._setgyear( (_mod(year,4)||(_mod(year,100)==0&&_mod(year,400)))? 1971 : 1972 );
	}
	else
	{
	    gtm._setgyear( (_mod(year,4)||(_mod(year,100)==0&&_mod(year,400)))? 1999 : 2000 );
	}

	tm = gtm.unixtime();
    }

    localtime_r( &tm, &locstm );

    return locstm.tm_isdst;
}

/**************************************************************************************************************/
long
gtimet::lutcoffset() const
{
    struct tm  locstm;
    time_t     tm( unixtime() );

    if ( tm < 0 )
    {
	gtimet	     gtm( *this );
	long double  year( gtm.gyear() );

	if ( _tmsl < TMSL19700101 )
	{
	    gtm._setgyear( (_mod(year,4)||(_mod(year,100)==0&&_mod(year,400)))? 1971 : 1972 );
	}
	else
	{
	    gtm._setgyear( (_mod(year,4)||(_mod(year,100)==0&&_mod(year,400)))? 1999 : 2000 );
	}

	tm = gtm.unixtime();
    }

    localtime_r( &tm, &locstm );

    return locstm.tm_gmtoff;
}

/**************************************************************************************************************/
ltimet&
ltimet::setlall( long double Y, int Mo, int D, int H, int Mi, long double S )
{
    long double  year( Y );
    int	         month( Mo );
    int		 mday( D );
    int		 hour( H );
    int		 minute( Mi );
    long double  second( S );
    
    if 
    (
    	month < 0 || month >= 12 ||
	mday < 0 || mday >= 28 ||
	hour < 0 || hour >= 24 ||
	minute < 0 || minute >= 60 ||
	second < 0.0L || second >= 59.0L
    )
    {
	gtimet  gtm( year, month, mday, hour, minute, second );
    
    	year = gtm.gyear();
	month = gtm.gmonth();
	mday = gtm.gmday();
	hour = gtm.ghour();
	minute = gtm.gminute();
	second = gtm.gsecond();
    }
    
    long double	 intsecond;
    long double	 fracsecond;

    fracsecond = modf( second, &intsecond );

    struct tm  locstm;
    int	       tmyear;

    if ( year < 1970 )
    {
	tmyear = (_mod(year,4)||(_mod(year,100)==0&&_mod(year,400)))? 71 : 72; // Year 1971:1972
    }
    else if ( year > numeric_limits<int>::max() )
    {
	tmyear = (_mod(year,4)||(_mod(year,100)==0&&_mod(year,400)))? 99 : 100; // Year 1999:2000
    }
    else
    {
    	tmyear = static_cast<int>( year ) - 1900;
    }

    locstm.tm_year = tmyear;
    locstm.tm_isdst = -1;
    
    locstm.tm_mon = month;
    locstm.tm_mday = mday + 1;
    locstm.tm_hour = hour;
    locstm.tm_min = minute;
    locstm.tm_sec = static_cast<int>( intsecond );
    
    struct tm  copystm( locstm );
    time_t     loctm( mktime(&locstm) );

    if ( loctm == static_cast<time_t>(-1) || copystm.tm_hour != locstm.tm_hour )
    {
    	// Keeping time linearity on spring localtime hole
    	locstm = copystm;
        locstm.tm_hour++;
	locstm.tm_min = 0;
        locstm.tm_sec = 0;
	loctm = mktime( &locstm );
	fracsecond = 0.0L;
    }
    
    setunixtime( loctm );
    _setgyear( year + gyear() - tmyear - 1900 );
    _tmsl += fracsecond;

    if ( lyear() != Y )
    {
	_distortionflags |= LYEARBIT;
    }

    if ( lmonth() != Mo )
    {
	_distortionflags |= LMONTHBIT;
    }

    if ( lmday() != D )
    {
	_distortionflags |= LMDAYBIT;
    }

    if ( lhour() != H )
    {
	_distortionflags |= LHOURBIT;
    }

    if ( lminute() != Mi )
    {
	_distortionflags |= LMINUTEBIT;
    }

    if ( lsecond() != S )
    {
	_distortionflags |= LSECONDBIT;
    }

    return *this;
}

/**************************************************************************************************************/
string
ltimet::str() const
/*
    ISO-8601 representation
*/
{
    ostringstream  oss;

    if ( isabsolutebegin() )
    {
	oss << "-infinity";
    }
    else if ( isabsoluteend() )
    {
	oss << "infinity";
    }
    else
    {
	oss << fixed << setprecision(0) << lyear()
	    << "-" << setfill('0') << setw(2) << lmonth()+1
	    << "-" << setfill('0') << setw(2) << lmday()+1
	    << " " << setfill('0') << setw(2) << lhour()
	    << ":" << setfill('0') << setw(2) << lminute()
	    << ":" << fixed << setprecision(0) << setfill('0') << setw(2) << floor(lsecond())
	    << showpos << setfill('0') << internal << setw(3) << ltzhour()
	    << noshowpos << setfill('0') << setw(2) << ltzminute();
    }

    return oss.str();
}
