/* tzinfo.cc
****************************************************************************************************************
****************************************************************************************************************

    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Askar Almatov

    This file is part of Libiclass. 
    
    Libiclass is free software: you can redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
    option) any later version.
    
    Libiclass is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.
    
    You should have received a copy of the GNU General Public License along with Libiclass. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <limits>
#include <libiclass/tzinfo>
#include <netinet/in.h>

using namespace iclass;
using namespace std;

const char*  TZFILEDEFAULT = "/etc/localtime";

/**************************************************************************************************************/
struct tzfileheadt 
{
    char  tzh_magic[4];		/* TZ_MAGIC */
    char  tzh_reserved[16];	/* reserved for future use */
    char  tzh_ttisgmtcnt[4];	/* coded number of trans. time flags */
    char  tzh_ttisstdcnt[4];	/* coded number of trans. time flags */
    char  tzh_leapcnt[4];	/* coded number of leap seconds */
    char  tzh_timecnt[4];	/* coded number of transition times */
    char  tzh_typecnt[4];	/* coded number of local time types */
    char  tzh_charcnt[4];	/* coded number of abbr. chars */
};
    
/**************************************************************************************************************/
struct tzfiletypet
{
    char	   tzt_offset[ 4 ];
    unsigned char  tzt_isdst;
    unsigned char  tzt_zonenameindex;
};

/**************************************************************************************************************/
struct tzfilelst
{
    char  tzl_time[ 4 ];
    char  tzl_correction[ 4 ];
};

/**************************************************************************************************************/
inline
tztransitiont::tztransitiont
(
    time_t  Timer, 
    long    Offset, 
    string  Zonename, 
    bool    Isdst, 
    bool    Isstd, 
    bool    Isgmt
) :
    _begin( Timer ),
    _offset( Offset ),
    _zonename( Zonename ),
    _isdst( Isdst ),
    _isstd( Isstd ),
    _isgmt( Isgmt )
{
}

/**************************************************************************************************************/
tzinfot::tzinfot()
{
    this->_read( TZFILEDEFAULT );
}

/**************************************************************************************************************/
void
tzinfot::_read( string Tzfile )
{
    ifstream     tzf( Tzfile.c_str() );
    tzfileheadt  tzh;

    tzf.read( reinterpret_cast<char*>(&tzh), sizeof(tzh) );
    
    int		    ttisgmtcnt( ntohl(*reinterpret_cast<uint32_t*>(tzh.tzh_ttisgmtcnt)) );
    int		    ttisstdcnt( ntohl(*reinterpret_cast<uint32_t*>(tzh.tzh_ttisstdcnt)) );
    int		    leapcnt( ntohl(*reinterpret_cast<uint32_t*>(tzh.tzh_leapcnt)) );
    int		    timecnt( ntohl(*reinterpret_cast<uint32_t*>(tzh.tzh_timecnt)) );
    int		    typecnt( ntohl(*reinterpret_cast<uint32_t*>(tzh.tzh_typecnt)) );
    int 	    charcnt( ntohl(*reinterpret_cast<uint32_t*>(tzh.tzh_charcnt)) );
    char*           ptzfttime( new char[4*timecnt] );
    unsigned char*  ptzfttypeindex( new unsigned char[timecnt] );
    tzfiletypet*    ptzfttype( new tzfiletypet[typecnt] );
    tzfilelst*      ptzflstransition( new tzfilelst[leapcnt] );
    char*	    ptzfisgmt( new char[ttisgmtcnt] );
    char*           ptzfisstd( new char[ttisstdcnt] );
    char*	    zonenames( new char[charcnt] );
    
    tzf.read( ptzfttime, 4*timecnt );
    tzf.read( reinterpret_cast<char*>(ptzfttypeindex), timecnt );
    tzf.read( reinterpret_cast<char*>(ptzfttype), typecnt*sizeof(tzfiletypet) );
    tzf.read( zonenames, charcnt );
    tzf.read( reinterpret_cast<char*>(ptzflstransition), leapcnt*sizeof(tzfilelst) );
    tzf.read( ptzfisstd, ttisstdcnt );
    tzf.read( ptzfisgmt, ttisgmtcnt );

    for ( int i=0; i<leapcnt; i++ )
    {
    	_lstransitions.push_back( ntohl(*reinterpret_cast<uint32_t*>(ptzflstransition[i].tzl_time)) );
	_lscorrections.push_back( ntohl(*reinterpret_cast<uint32_t*>(ptzflstransition[i].tzl_correction)) );
    }
    
    tztransitiont  firsttrn
    (
    	numeric_limits< time_t >::min(),
	ntohl( *reinterpret_cast<uint32_t*>(ptzfttype[0].tzt_offset) ),
	&zonenames[ ptzfttype[0].tzt_zonenameindex ],
	ptzfttype[ 0 ].tzt_isdst,
	ptzfisstd[ 0 ],
	ptzfisgmt[ 0 ]
    );

    _trnarray.push_back( firsttrn );
    
    for ( int i=0; i<timecnt; i++ )
    {
        int  typeindex( ptzfttypeindex[i] );
	
    	tztransitiont  nexttrn
	(
	    ntohl( *reinterpret_cast<uint32_t*>(&ptzfttime[4*i]) ),
	    ntohl( *reinterpret_cast<uint32_t*>(ptzfttype[typeindex].tzt_offset) ),
	    &zonenames[ ptzfttype[typeindex].tzt_zonenameindex ],
	    ptzfttype[ typeindex ].tzt_isdst,
	    ptzfisstd[ typeindex ],
	    ptzfisgmt[ typeindex ]
	);
	
        _trnarray.push_back( nexttrn );
    }
    
    delete[] ptzfttime;
    delete[] ptzfttypeindex;
    delete[] ptzfttype;
    delete[] ptzflstransition;
    delete[] ptzfisstd;
    delete[] ptzfisgmt;
    delete[] zonenames;
}

/**************************************************************************************************************/
const tztransitiont&
tzinfot::findasgmt( time_t Timer ) const
{
    vector< tztransitiont >::const_iterator  vi( lower_bound(_trnarray.begin(), _trnarray.end(), Timer) );

    return *--vi;
}

